
!==============================================================================!
!==============================================================================!
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!==============================================================================!
!==============================================================================!
!                                                                              |
!              TRACKING CLOUD POST PROCESSING   Jan. 2015                      |
!              written by Matthias Schmiddunser                                |
!              using age calculation form Pecube and Terra                     |
!                                                                              |
!==============================================================================!
!==============================================================================!
!                                                                              |
! This routine reads in the tracking output files, extracts the surfaced       |
! tracks, does variable calculation and writes ASCII vtk files.                |
!                                                                              |
! CAUTION: Particle ID (Douar) and Point Number (VTK) differ: DOUAR numbering  |
!     starts at 1, VTK numbering starts at 0.                                  |

program post_tracking

use trdefinitions
use version
use m_mad_he
use m_mad_trax
use m_mad_trax_zirkon
use m_age_algorithms
use m_toolbox
use m_write_tecplot

implicit none

include 'mpif.h'

type (tracking_cloud) :: trcl
type (input_parameters) :: inparam
type (output_parameters) :: outparam
type (trsheet), dimension(:), allocatable :: surface
 
!buffer variables for reading
real :: read_time,lref,veloref,rhoref,tempscale
integer :: read_step,read_ntracks
integer,dimension(:,:),allocatable :: read_tracks
integer,dimension(:),allocatable :: keep_tracks

!mpi and data transfer variables
logical :: flag
integer :: provided,iproc,nproc,ierr
integer (kind=MPI_ADDRESS_KIND) :: maxtag
integer,dimension(MPI_STATUS_SIZE) :: mpistatus
integer :: sender,tracksent,send_nstep,endstep
real,dimension(:),allocatable :: send_time,send_temp
real,dimension(:),allocatable :: allages

!age calculation
real(8),DIMENSION(21) :: Mad_He_ages
real(8) :: Mad_AHe_std
real(8) :: Mad_AHe_small
real(8) :: Mad_AHe_medium
real(8) :: Mad_AHe_large
real(8) :: Mad_AHe_lowdam
real(8) :: Mad_AHe_moddam
real(8) :: Mad_AHe_hidam
real(8) :: Mad_ZHe
real(8) :: Mad_ArKFs
real(8) :: Mad_ArBio
real(8) :: Mad_ArHorn
real(8) :: Mad_ArMusc
real(8) :: Mad_Ti_UTh_He
real(8) :: Mad_AFT
real(4) :: Mad_ZFT
real(8) :: age_Muscovite
real(8) :: age_ZFT
real(8) :: age_Apatite_U_Th_Pb
real(8) :: age_Biotite
real(8) :: age_Rutile_U_Pb
real(8) :: age_Titanite_U_Pb
real(8) :: age_Zircon_U_Pb    

integer, DIMENSION(21) :: Mad_He_inputs

real, dimension (:,:), allocatable :: origpos
real :: xorigmin, xorigmax, yorigmin, yorigmax, zorigmin, zorigmax

!other variables
integer :: i,k,l,err,dummyint,itrack,nbins
real :: dummyreal
double precision :: dummydble
double precision,dimension(17) :: dummydble17
logical :: keepme,versionconflict

character(4) :: cistep

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

provided=0
call mpi_init_thread(MPI_THREAD_MULTIPLE,provided,ierr)
call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

if (iproc==0) then
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,'(a)') '!                                                                              |'
  write (*,'(a)') '!          8888888b.   .d88888b.  888     888       d8888 8888888b.            |'
  write (*,'(a)') '!          888  "Y88b d88P" "Y88b 888     888      d88888 888   Y88b           |'
  write (*,'(a)') '!          888    888 888     888 888     888     d88P888 888    888           |'
  write (*,'(a)') '!          888    888 888     888 888     888    d88P 888 888   d88P           |'
  write (*,'(a)') '!          888    888 888     888 888     888   d88P  888 8888888P"            |'
  write (*,'(a)') '!          888    888 888     888 888     888  d88P   888 888 T88b             |'
  write (*,'(a)') '!          888  .d88P Y88b. .d88P Y88b. .d88P d8888888888 888  T88b            |'
  write (*,'(a)') '!          8888888P"   "Y88888P"   "Y88888P" d88P     888 888   T88b           |'
  write (*,'(a)') '!                                                                              |'
  write (*,'(a)') '!             DOUAR-WSMP post-processor for the tracking cloud                 |'
write (*,'(a,i1,a,i1,a,i1,a,i1,a)') '!                                      DOUAR version ', & 
    versionmajor,'.',versionminor,'.',versionstat,'.',versionrev,'                   |'
  write (*,'(a,i3,a)') '!             Running on ', nproc,' nodes                                             |'
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,*) ''
end if

!------------------------------------------------------------------------------|
!    Input parameters
!------------------------------------------------------------------------------|

!Reading input parameters from post-input-file:

versionconflict = .false.
if (iproc==0) then
    open(unit=10, file='input_tracking_outputs.txt',status='old')

    read (10,*) inparam%vermajor
    read (10,*) inparam%verminor
    read (10,*) inparam%verstat
    if (versionmajor/=inparam%vermajor .or. versionminor/=inparam%verminor & 
          .or. versionstat/=inparam%verstat) then
        write (*,'(a)') 'Warning: Post-processor version and post input file'// & 
            ' version do not agree.'
        write (*,'(a,i1,a,i1,a,i1)') 'Post-processor version: ', &
            versionmajor,'.',versionminor,'.',versionstat
        write (*,'(a,i1,a,i1,a,i1)') 'Post input file version: ' &
                ,inparam%vermajor,'.',inparam%verminor,'.',inparam%verstat
        !write (*,'(a)') 'Unable to continue. Exiting.'
        versionconflict = .true.
        !call mpi_bcast (versionconflict,1,mpi_logical,0,mpi_comm_world,ierr)
        !call stop_run ('Input file version conflict')
    endif
    read (10,*) inparam%dir
    read (10,*) outparam%dir
    read (10,*) outparam%mode
    read (10,*) outparam%outformat
    read (10,*) outparam%scaled
    read (10,*) outparam%first
    read (10,*) outparam%last
    read (10,*) outparam%step
    read (10,*) outparam%keep_nstep
    read (10,*) outparam%min_track
    read (10,*) nbins
    !outparam%keep_nstep has to be less than outparam%step
    outparam%keep_nstep = min(outparam%keep_nstep,outparam%step-1)
    read (10,*) inparam%first
    read (10,*) inparam%step
    read (10,*)  !here come the age parameters
    read (10,*) outparam%pad_age
    read (10,*) outparam%Mad_AHe_std
    read (10,*) outparam%Mad_AHe_small
    read (10,*) outparam%Mad_AHe_medium
    read (10,*) outparam%Mad_AHe_large
    read (10,*) outparam%Mad_AHe_lowdam
    read (10,*) outparam%Mad_AHe_moddam
    read (10,*) outparam%Mad_AHe_hidam
    read (10,*) outparam%Mad_ZHe
    read (10,*) outparam%Mad_ArKFs
    read (10,*) outparam%Mad_ArBio
    read (10,*) outparam%Mad_ArHorn
    read (10,*) outparam%Mad_ArMusc
    read (10,*) outparam%Mad_Ti_UTh_He
    read (10,*) outparam%Mad_AFT
    read (10,*) outparam%Mad_AFT_param
    read (10,*) outparam%Mad_ZFT
    read (10,*) outparam%Mad_ZFT_param
    read (10,*) outparam%Muscovite
    read (10,*) outparam%ZFT
    read (10,*) outparam%Apatite_U_Th_Pb
    read (10,*) outparam%Biotite
    read (10,*) outparam%Rutile_U_Pb
    read (10,*) outparam%Titanite_U_Pb
    read (10,*) outparam%Zircon_U_Pb    
    read (10,*) outparam%origin
    read (10,*) outparam%maxdepth
    read (10,*) outparam%maxtemp
    read (10,*) outparam%maxpressure
    read (10,*) outparam%ageID
    read (10,*)  !here come the track parameters
    read (10,*) outparam%velo
    read (10,*) outparam%veloz
    read (10,*) outparam%temp
    read (10,*) outparam%press
    read (10,*) outparam%time
    read (10,*) outparam%istep
    read (10,*) outparam%trackID

    close (10)

    !calculating and andjusting input and output timesteps
    outparam%nstep = (outparam%last-outparam%first)/outparam%step+1
    outparam%last = outparam%first+(outparam%nstep-1)*outparam%step

    inparam%last = outparam%last
    inparam%nstep = (inparam%last-inparam%first)/inparam%step+1
    inparam%last = inparam%first+(inparam%nstep-1)*inparam%step

    !Checking for slash at end of input and output dir
    l = len(trim(inparam%dir))
    if (inparam%dir(l:l) .ne. '/') then
        inparam%dir = inparam%dir(1:l)//'/'
    end if
    l = len(trim(outparam%dir))
    if (outparam%dir(l:l) .ne. '/') then
        outparam%dir = outparam%dir(1:l)//'/'
    end if
else
    !call mpi_bcast (versionconflict,1,mpi_logical,0,mpi_comm_world,ierr)
    !if (versionconflict) call stop_run ('Input file version conflict')
end if

!Distribute output parameters to other processes
call mpi_bcast (outparam%mode,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%pad_age,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AHe_std,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AHe_small,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AHe_medium,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AHe_large,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AHe_lowdam,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AHe_moddam,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AHe_hidam,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_ZHe,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_ArKFs,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_ArBio,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_ArMusc,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_ArHorn,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_Ti_UTh_He,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AFT,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_AFT_param,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_ZFT,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Mad_ZFT_param,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Muscovite,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%ZFT,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Apatite_U_Th_Pb,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Biotite,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Rutile_U_Pb,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Titanite_U_Pb,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%Zircon_U_Pb,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%origin,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%maxdepth,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%maxtemp,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%maxpressure,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%ageID,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%velo,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%veloz,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%temp,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%press,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%time,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%istep,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast (outparam%trackID,1,mpi_integer,0,mpi_comm_world,ierr)

!crate input flag array for Mad_He:

Mad_He_inputs = 0
if (outparam%Mad_AHe_std==1) Mad_He_inputs(1) = 1
if (outparam%Mad_AHe_small==1) Mad_He_inputs(2) = 1
if (outparam%Mad_AHe_medium==1) Mad_He_inputs(3) = 1
if (outparam%Mad_AHe_large==1) Mad_He_inputs(4) = 1
if (outparam%Mad_AHe_lowdam==1) Mad_He_inputs(5) = 1
if (outparam%Mad_AHe_moddam==1) Mad_He_inputs(6) = 1
if (outparam%Mad_AHe_hidam==1) Mad_He_inputs(7) = 1
if (outparam%Mad_ZHe==1) Mad_He_inputs(8) = 1
if (outparam%Mad_ArKFs==1) Mad_He_inputs(12) = 1
if (outparam%Mad_ArBio==1) Mad_He_inputs(13) = 1
if (outparam%Mad_ArMusc==1) Mad_He_inputs(14) = 1
if (outparam%Mad_ArHorn==1) Mad_He_inputs(15) = 1
if (outparam%Mad_Ti_UTh_He==1) Mad_He_inputs(21) = 1


!determine number of ages that will be calculated

outparam%nages = sum(Mad_He_inputs) + & !Mad_He inputs
    outparam%Mad_AFT +  & !Mad_Trax inputs
    outparam%Mad_ZFT +  & !Mad_Trax_zirc inputs
    outparam%Muscovite + outparam%ZFT + outparam%Apatite_U_Th_Pb + &
    outparam%Biotite + outparam%Rutile_U_Pb + outparam%Titanite_U_Pb + &
    outparam%Zircon_U_Pb !age_algorithms inputs

!write (* ,*) 'iproc',iproc , 'nages:', outparam%nages

allocate (allages(outparam%nages),stat=err); if (err.ne.0) call stop_run ('Error alloc allages')

!writing output parameters to screen
if (iproc==0) then
    write (*,'(a,a)') 'Simulation Directory: ',inparam%dir
    write (*,'(a,i4,a,i4,a,i4,a,i4)') 'Input Timesteps: Reading ',inparam%nstep,' input files '// & 
    'from step ',inparam%first,' to ',inparam%last,' in steps of ',inparam%step
    write (*,'(a,a)') 'Output Directory: ',outparam%dir
    write (*,'(a,i4,a,i4,a,i4,a,i4)') 'Output Timesteps: Generating ',outparam%nstep,' output files '// &
    'from step ',outparam%first,' to ',outparam%last,' in steps of ',outparam%step
    write (*,'(a,i2,a)') 'accumulating ',outparam%keep_nstep,' steps for surfaced tracks'
    write (*,'(a,i1)') 'Output mode: ',outparam%mode
    write (*,'(a,l)') 'Write natural units:     ',(outparam%scaled==1)
    write (*,'(a)') ''
    write (*,'(a)') ' Age variables: '
    write (*,'(a,l)') 'Write Mad AHe std grain       ',(outparam%Mad_AHe_std==1)
    write (*,'(a,l)') 'Write Mad AHe small grain     ',(outparam%Mad_AHe_small==1)
    write (*,'(a,l)') 'Write Mad AHe medium grain    ',(outparam%Mad_AHe_medium==1)
    write (*,'(a,l)') 'Write Mad AHe large grain     ',(outparam%Mad_AHe_large==1)
    write (*,'(a,l)') 'Write Mad AHe low damage      ',(outparam%Mad_AHe_lowdam==1)
    write (*,'(a,l)') 'Write Mad AHe moderate damage ',(outparam%Mad_AHe_moddam==1)
    write (*,'(a,l)') 'Write Mad AHe high damage     ',(outparam%Mad_AHe_hidam==1)
    write (*,'(a,l)') 'Write Mad ZHe                 ',(outparam%Mad_ZHe==1)
    write (*,'(a,l)') 'Write Mad Ar in K-Feldspar    ',(outparam%Mad_ArKFs==1)
    write (*,'(a,l)') 'Write Mad Ar in Biotite       ',(outparam%Mad_ArBio==1)
    write (*,'(a,l)') 'Write Mad Ar in Muscovite     ',(outparam%Mad_ArMusc==1)
    write (*,'(a,l)') 'Write Mad Ar in Hornblende    ',(outparam%Mad_ArHorn==1)
    write (*,'(a,l)') 'Write Mad Titanite (U-Th)/He  ',(outparam%Mad_Ti_UTh_He==1)
    write (*,'(a,l)') 'Write Mad AFT                 ',(outparam%Mad_AFT==1)
    write (*,'(a,l)') 'Mad AFT parameter             ',(outparam%Mad_AFT_param)
    write (*,'(a,l)') 'Write Mad ZFT                 ',(outparam%Mad_ZFT==1)
    write (*,'(a,l)') 'Mad ZFT parameter             ',(outparam%Mad_ZFT_param)
    write (*,'(a,l)') 'Write Muscovite               ',(outparam%Muscovite==1)
    write (*,'(a,l)') 'Write ZFT                     ',(outparam%ZFT==1)
    write (*,'(a,l)') 'Write Apatite (U-Th)/Pb       ',(outparam%Apatite_U_Th_Pb==1)
    write (*,'(a,l)') 'Write Biotite                 ',(outparam%Biotite==1)
    write (*,'(a,l)') 'Write Rutile U/Pb             ',(outparam%Rutile_U_Pb==1)
    write (*,'(a,l)') 'Write Titanite U/Pb           ',(outparam%Titanite_U_Pb==1)
    write (*,'(a,l)') 'Write Zircon U/Pb             ',(outparam%Zircon_U_Pb==1)     
    write (*,'(a,l)') 'Write particle origin    ',(outparam%origin==1)
    write (*,'(a,l)') 'Write maximum depth      ',(outparam%maxdepth==1)
    write (*,'(a,l)') 'Write track ID           ',(outparam%ageID==1)
    write (*,'(a)') ''
    write (*,'(a)') 'Track variables: '
    write (*,'(a,l)') 'Write track velocity:    ',(outparam%velo==1)
    write (*,'(a,l)') 'Write track z velocity:  ',(outparam%veloz==1)
    write (*,'(a,l)') 'Write track temperature: ',(outparam%temp==1)
    write (*,'(a,l)') 'Write track pressure:    ',(outparam%press==1)
    write (*,'(a,l)') 'Write track time:        ',(outparam%time==1)
    write (*,'(a,l)') 'Write track step number: ',(outparam%istep==1)
    write (*,'(a,l)') 'Write particle/track ID: ',(outparam%trackID==1)
    write (*,'(a)') ''
    write (*,'(a)') '!------------------------------------------------------------------------------|'
    write (*,'(a)') ''
end if

!------------------------------------------------------------------------------|
!    Reading Tracking Cloud Files
!------------------------------------------------------------------------------|

if (iproc==0) then
    !reading input file header for array sizes
    call int_to_char (cistep,4,inparam%last)

    open (unit=20,file=trim(inparam%dir)//'TROUT/tracking_t'//cistep//'.bin', &
           status='old',form='unformatted')
    read (20) trcl%np,read_ntracks,read_step,read_time
    read (20) lref,veloref,rhoref,tempscale
    close (20)

    !allocating master arrays for input
    allocate (trcl%x(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%x')
    allocate (trcl%y(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%y')
    allocate (trcl%z(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%z')
    allocate (trcl%temp(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%temp')
    allocate (trcl%press(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%press')
    allocate (trcl%step(inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%step')
    allocate (trcl%time(inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%time')
    allocate (read_tracks(read_ntracks,3),stat=err)
    if (err.ne.0) call stop_run ('Error alloc read_tracks')

    !read all input files and save in collective arrays
    k = 1
    do i = inparam%first,inparam%last,inparam%step
        !read all information for this time step
        call int_to_char (cistep,4,i)
        open (unit=20,file=trim(inparam%dir)//'TROUT/tracking_t'//cistep//'.bin', & 
               status='old',form='unformatted')
        
        read (20) dummyint,dummyint,trcl%step(k),trcl%time(k)
        read (20) 
        read (20) trcl%x(:,k)
        read (20) trcl%y(:,k)
        read (20) trcl%z(:,k)
        read (20) trcl%temp(:,k)
        read (20) trcl%press(:,k)
        
        !read track information only from last timestep
        if (i==inparam%last) then
            read (20) read_tracks
            write (*,'( a)') 'All surfaced tracks:'
            write (*,'(a)') 'track ID, Particle ID,  start,    end'
            do l=1,read_ntracks 
                write (*,'(i8,i13,2i8)') l, read_tracks(l,:)
            end do
            write (*,'(a)') ''
        end if

        !advance column in master array
        k = k+1

        close (20)
    end do

    write (*,'(a)') 'Finished reading input files'
    write (*,'(a)') ''

    !select only tracks that surface at an output timestep or up to outparam%keep_nstep
    !timesteps before each output timestep and are at least min_track_length 
    !timesteps long.
    !Then adjust track timesteps to actual position in master array.
    !The offset k has to be recorded in order to associate the track with the
    !correct timestep.

    allocate (keep_tracks(read_ntracks),stat=err)
    if (err.ne.0) call stop_run ('Error alloc keep_tracks')
    keep_tracks = 0
    trcl%ntracks = 0
    do k=0,outparam%keep_nstep
        do i=1,read_ntracks
            keepme = read_tracks(i,3) >= outparam%first-k
            keepme = keepme .AND. (mod((read_tracks(i,3)+k-outparam%first),outparam%step)==0)
            keepme = keepme .AND. read_tracks(i,3)-read_tracks(i,2)+1 >= outparam%min_track
            if (keepme) then
                keep_tracks(i) = 1000+k
                trcl%ntracks = trcl%ntracks+1
            end if
        end do
    end do

    !allocate tracks array of needed length

    if (trcl%ntracks > 0) then
        allocate (trcl%tracks(trcl%ntracks,7),stat=err)
        if (err.ne.0) call stop_run ('Error alloc trcl%tracks')
        allocate (origpos(trcl%ntracks,3),stat=err)
        if (err.ne.0) call stop_run ('Error alloc origpos')
        trcl%tracks = 0

        !copy relevant tracks into new track array
        !adjust track timsteps to position in new track array
        !always round up for start and round down for end in order not to 
        !include parts of other tracks
        itrack=1
        write (*,'(a)') 'Selected surfaced tracks:'
        write (*,'(a)') '*: in master array'
        write (*,'(a)') 'Particle ID, start*,   end*, length*, output timestep, output step*, track ID'
        do i=1,read_ntracks
            if (keep_tracks(i)>0) then
                !1st column is particle ID
                trcl%tracks(itrack,1) = read_tracks(i,1)
                !2nd column is start in master array, round up
                trcl%tracks(itrack,2)= max(1,ceiling(1.0+(read_tracks(i,2)-inparam%first)/(1.0*inparam%step)))
                !3rd columd is end in master array, round down
                trcl%tracks(itrack,3)= max(1,1+(read_tracks(i,3)-inparam%first)/inparam%step) !integer division
                !4th column is track length (number of points in master array)
                trcl%tracks(itrack,4) = trcl%tracks(itrack,3)-trcl%tracks(itrack,2)+1
                !5th column is Douar timestep of surfacing, upshifted for those tracks 
                !within outparam%keep_nstep
                trcl%tracks(itrack,5) = read_tracks(i,3)+(keep_tracks(i)-1000)
                !6th column is output timestep in master array
                trcl%tracks(itrack,6)= max(1,1+(trcl%tracks(itrack,5)-inparam%first)/inparam%step) !integer division
                !7th column is track ID, i.e. the row of the original track array from input file
                trcl%tracks(itrack,7) = i

                origpos(itrack,1) = trcl%x(trcl%tracks(itrack,1), trcl%tracks(itrack,2))
                origpos(itrack,2) = trcl%y(trcl%tracks(itrack,1), trcl%tracks(itrack,2))
                origpos(itrack,3) = trcl%z(trcl%tracks(itrack,1), trcl%tracks(itrack,2))

                write (*,'(i11,i8,i8,i9,i17,i14,i10)') trcl%tracks(itrack,:)
                itrack = itrack+1
            end if
        end do

        !Calculate orign box off all surfaced tracks
        xorigmin = minval( origpos(:,1) )
        xorigmax = maxval( origpos(:,1) )
        yorigmin = minval( origpos(:,2) )
        yorigmax = maxval( origpos(:,2) )
        zorigmin = minval( origpos(:,3) )
        zorigmax = maxval( origpos(:,3) )

        write (*,*) "Surfaced particles originate from this box:"
        if (outparam%scaled==1) then
            write (*,'(a,f12.5,a,f12.5)') 'x: ', xorigmin*lref, ' to ', xorigmax*lref
            write (*,'(a,f12.5,a,f12.5)') 'y: ', yorigmin*lref, ' to ', yorigmax*lref
            write (*,'(a,f12.5,a,f12.5)') 'z: ', zorigmin*lref, ' to ', zorigmax*lref
        else
            write (*,'(a,f12.5,a,f12.7)') 'x: ', xorigmin, ' to ', xorigmax
            write (*,'(a,f12.5,a,f12.7)') 'y: ', yorigmin, ' to ', yorigmax
            write (*,'(a,f12.5,a,f12.7)') 'z: ', zorigmin, ' to ', zorigmax
        end if
        deallocate(origpos)
    else
        write (*,'(a)') 'No surfaced tracks'
    end if

    deallocate (read_tracks)
    deallocate (keep_tracks)

end if

call mpi_bcast(trcl%ntracks,1,mpi_integer,0,mpi_comm_world,ierr)

!------------------------------------------------------------------------------|
!    Calculate Thermochron Ages
!------------------------------------------------------------------------------|
ages: if ((btest(outparam%mode, 0)) .AND. trcl%ntracks>0) then
    multiproc: if (nproc > 1) then
        !Check for enough tags for self scheduling
        call MPI_COMM_GET_ATTR (MPI_COMM_WORLD, MPI_TAG_UB, maxtag, flag, ierr)
        if (int(trcl%ntracks,MPI_ADDRESS_KIND) > maxtag) then
            call stop_run ('Too many tracks for safe communications')
        end if

        workshare: if (iproc==0) then
        !master distributing work
            if (outparam%Mad_AHe_std==1)     allocate (trcl%Mad_AHe_std(trcl%ntracks))
            if (outparam%Mad_AHe_small==1)   allocate (trcl%Mad_AHe_small(trcl%ntracks))
            if (outparam%Mad_AHe_medium==1)  allocate (trcl%Mad_AHe_medium(trcl%ntracks))
            if (outparam%Mad_AHe_large==1)   allocate (trcl%Mad_AHe_large(trcl%ntracks))
            if (outparam%Mad_AHe_lowdam==1)  allocate (trcl%Mad_AHe_lowdam(trcl%ntracks))
            if (outparam%Mad_AHe_moddam==1)  allocate (trcl%Mad_AHe_moddam(trcl%ntracks))
            if (outparam%Mad_AHe_hidam==1)   allocate (trcl%Mad_AHe_hidam(trcl%ntracks))
            if (outparam%Mad_ZHe==1)         allocate (trcl%Mad_ZHe(trcl%ntracks))
            if (outparam%Mad_ArKFs==1)       allocate (trcl%Mad_ArKFs(trcl%ntracks))
            if (outparam%Mad_ArBio==1)       allocate (trcl%Mad_ArBio(trcl%ntracks))
            if (outparam%Mad_ArMusc==1)      allocate (trcl%Mad_ArMusc(trcl%ntracks))
            if (outparam%Mad_ArHorn==1)      allocate (trcl%Mad_ArHorn(trcl%ntracks))
            if (outparam%Mad_Ti_UTh_He==1)   allocate (trcl%Mad_Ti_UTh_He(trcl%ntracks))
            if (outparam%Mad_AFT==1)         allocate (trcl%Mad_AFT(trcl%ntracks))
            if (outparam%Mad_ZFT==1)         allocate (trcl%Mad_ZFT(trcl%ntracks))
            if (outparam%Muscovite==1)       allocate (trcl%Muscovite(trcl%ntracks))
            if (outparam%ZFT==1)             allocate (trcl%ZFT(trcl%ntracks))
            if (outparam%Apatite_U_Th_Pb==1) allocate (trcl%Apatite_U_Th_Pb(trcl%ntracks))
            if (outparam%Biotite==1)         allocate (trcl%Biotite(trcl%ntracks))
            if (outparam%Rutile_U_Pb==1)     allocate (trcl%Rutile_U_Pb(trcl%ntracks))
            if (outparam%Titanite_U_Pb==1)   allocate (trcl%Titanite_U_Pb(trcl%ntracks))
            if (outparam%Zircon_U_Pb==1)     allocate (trcl%Zircon_U_Pb(trcl%ntracks))     

            tracksent = 0
            !write (* ,*) iproc, ': ntracks:', trcl%ntracks, 'nproc: ',nproc
            !wait for answer, distribute new track
            workdistr: do i=2-nproc,trcl%ntracks
                !for f irst  nproc-1 times, send to processes without receive
                !distribute one track to each processor

                if (i>0) then
                    !get re sult from other process
                    
                    call MPI_RECV (allages,outparam%nages,MPI_REAL,MPI_ANY_SOURCE, & 
                                    MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)
                    !Identify process and track ID
                    sender = mpistatus(MPI_SOURCE)
                    itrack = mpistatus(MPI_TAG)

                    !write (*,'(i3,a,i5,a,i3,a,i2)') iproc, ': received track',itrack,' from proc ',sender, 'length = ',outparam%nages

                    !Write into result arrays
                    k = 1
                    if (outparam%Mad_AHe_std==1) then
                        trcl%Mad_AHe_std(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_small==1) then
                        trcl%Mad_AHe_small(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_medium==1) then
                        trcl%Mad_AHe_medium(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_large==1) then
                        trcl%Mad_AHe_large(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_lowdam==1) then
                        trcl%Mad_AHe_lowdam(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_moddam==1) then
                        trcl%Mad_AHe_moddam(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_hidam==1) then
                        trcl%Mad_AHe_hidam(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_ZHe==1) then
                        trcl%Mad_ZHe(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_ArKFs==1) then
                        trcl%Mad_ArKFs(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_ArBio==1) then
                        trcl%Mad_ArBio(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_ArMusc==1) then
                        trcl%Mad_ArMusc(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_ArHorn==1) then
                        trcl%Mad_ArHorn(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_Ti_UTh_He==1) then
                        trcl%Mad_Ti_UTh_He(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_AFT==1) then
                        trcl%Mad_AFT(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Mad_ZFT==1) then
                        trcl%Mad_ZFT(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Muscovite==1) then
                        trcl%Muscovite(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%ZFT==1) then
                        trcl%ZFT(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Apatite_U_Th_Pb==1) then
                        trcl%Apatite_U_Th_Pb(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Biotite==1) then
                        trcl%Biotite(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Rutile_U_Pb==1) then
                        trcl%Rutile_U_Pb(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Titanite_U_Pb==1) then
                        trcl%Titanite_U_Pb(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Zircon_U_Pb==1) then
                        trcl%Zircon_U_Pb(itrack) = allages(k)
                        k = k+1
                    end if 
                else
                    !effecively loops from 1 to nproc-1 for sender
                    sender = i+nproc-1
                end if

                !Send new information
                if (tracksent < trcl%ntracks) then
                    !Length of time and temperature arrays 
                    !Add two entrys at front for initalisation
                    !If surfaced earlier than output timestep, add another entry at end
                    endstep = min(1,trcl%tracks(tracksent+1,6)-trcl%tracks(tracksent+1,3)) 
                    send_nstep = 2 + trcl%tracks(tracksent+1,4)+endstep

                    allocate (send_time(send_nstep))
                    allocate (send_temp(send_nstep))

                    !pad send-arrays at beginning, part 1
                    send_time(3:send_nstep-endstep) = &
                        trcl%time(trcl%tracks(tracksent+1,2):trcl%tracks(tracksent+1,3))
                    send_time(1:2) = (/0.0,0.0/)

                    send_temp(3:send_nstep-endstep) = trcl%temp(trcl%tracks(tracksent+1,1), &
                        trcl%tracks(tracksent+1,2):trcl%tracks(tracksent+1,3))
                    send_temp(1) = send_temp(3)
                    send_temp(2) = send_temp(3)
                    
                    !pad send arrays at end if necessary
                    if (endstep>0) then
                        send_time(send_nstep) = trcl%time(trcl%tracks(tracksent+1,6))
                        send_temp(send_nstep) = send_temp(send_nstep-1)
                    end if

                    do l=1,send_nstep
                        !Calculate relative Age in Ma
                        send_time(l) = send_time(send_nstep) - send_time(l)
                        send_time(l) = send_time(l)*lref/veloref 
                        !scale temperature
                        send_temp(l) = send_temp(l)*tempscale
                    end do

                    !pad age array at the beginning, part 2
                    send_time(1) = outparam%pad_age
                    send_time(2) = 0.5*(send_time(1)+send_time(3))

                    !write (*,*) iproc, ': send track',tracksent+1,' of length',send_nstep,' to proc ',sender
                    call MPI_SEND (send_time,send_nstep,MPI_REAL,sender,tracksent+1, & 
                                    MPI_COMM_WORLD,ierr)
                    call MPI_SEND (send_temp,send_nstep,MPI_REAL,sender,tracksent+1, & 
                                    MPI_COMM_WORLD,ierr)
                    tracksent = tracksent+1

                    deallocate (send_time)
                    deallocate (send_temp)

                else
                    !send end signal
                    call MPI_SEND (0.0,1,MPI_REAL,sender,0,MPI_COMM_WORLD,ierr)
                end if
            end do workdistr


        else workshare
        !slaves doing actual calculation
            !write (* ,*) 'iproc',iproc , 'ntracks:', trcl%ntracks, 'nproc: ',nproc
            agecalc: do
                call MPI_probe  (0, MPI_ANY_TAG, MPI_COMM_WORLD, mpistatus, ierr)
                itrack = mpistatus(MPI_TAG)
                !write (*,*) iproc, ': probed track', itrack

                !tag 0 is end signal
                if (itrack==0) then
                    !write (*,*) iproc, ': receive end signal'
                    !empty send buffer - necessary?
                    call MPI_RECV (dummyreal,1,MPI_REAL,0,MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)
                    exit
                end if

                call MPI_get_count (mpistatus,MPI_REAL,send_nstep,ierr)

                allocate (send_time(send_nstep))
                allocate (send_temp(send_nstep))

                !write (*,*) iproc, ': receiving track', itrack, ' of length', send_nstep
                call MPI_RECV (send_time,send_nstep,MPI_REAL,0, & 
                                MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)
                call MPI_RECV (send_temp,send_nstep,MPI_REAL,0, & 
                                MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)

                !do age calculation
                k=1
                !routines from Mad_He
                if (sum(Mad_He_inputs)>0) then
                    !write (*,*) 'Proc', iproc, 'Calling Mad_He for track ', itrack
                    call Mad_He (dble(send_time), dble(send_temp), send_nstep, Mad_He_ages,Mad_He_inputs,21)
                    if (outparam%Mad_AHe_std==1) then
                        Mad_AHe_std = Mad_He_ages(1)
                        call correct_number(Mad_AHe_std)
                        allages(k) = real(Mad_AHe_std)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_small==1) then
                        Mad_AHe_small = Mad_He_ages(2)
                        call correct_number(Mad_AHe_small)
                        allages(k) = real(Mad_AHe_small)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_medium==1) then
                        Mad_AHe_medium = Mad_He_ages(3)
                        call correct_number(Mad_AHe_medium)
                        allages(k) = real(Mad_AHe_medium)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_large==1) then
                        Mad_AHe_large = Mad_He_ages(4)
                        call correct_number(Mad_AHe_large)
                        allages(k) = real(Mad_AHe_large)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_lowdam==1) then
                        Mad_AHe_lowdam = Mad_He_ages(5)
                        call correct_number(Mad_AHe_lowdam)
                        allages(k) = real(Mad_AHe_lowdam)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_moddam==1) then
                        Mad_AHe_moddam = Mad_He_ages(6)
                        call correct_number(Mad_AHe_moddam)
                        allages(k) = real(Mad_AHe_moddam)
                        k = k+1
                    end if
                    if (outparam%Mad_AHe_hidam==1) then
                        Mad_AHe_hidam = Mad_He_ages(7)
                        call correct_number(Mad_AHe_hidam)
                        allages(k) = real(Mad_AHe_hidam)
                        k = k+1
                    end if
                    if (outparam%Mad_ZHe==1) then
                        Mad_ZHe = Mad_He_ages(8)
                        call correct_number(Mad_ZHe)
                        allages(k) = real(Mad_ZHe)
                        k = k+1
                    end if
                    if (outparam%Mad_ArKFs==1) then
                        Mad_ArKFs = Mad_He_ages(12)
                        call correct_number(Mad_ArKFs)
                        allages(k) = real(Mad_ArKFs)
                        k = k+1
                    end if
                    if (outparam%Mad_ArBio==1) then
                        Mad_ArBio = Mad_He_ages(13)
                        call correct_number(Mad_ArBio)
                        allages(k) = real(Mad_ArBio)
                        k = k+1
                    end if
                    if (outparam%Mad_ArMusc==1) then
                        Mad_ArMusc = Mad_He_ages(14)
                        call correct_number(Mad_ArMusc)
                        allages(k) = real(Mad_ArMusc)
                        k = k+1
                    end if
                    if (outparam%Mad_ArHorn==1) then
                        Mad_ArHorn = Mad_He_ages(15)
                        call correct_number(Mad_ArHorn)
                        allages(k) = real(Mad_ArHorn)
                        k = k+1
                    end if
                    if (outparam%Mad_Ti_UTh_He==1) then
                        Mad_Ti_UTh_He = Mad_He_ages(21)
                        call correct_number(Mad_Ti_UTh_He)
                        allages(k) = real(Mad_Ti_UTh_He)
                        k = k+1
                    end if
                end if
                !routines from Mad_Trax
                if (outparam%Mad_AFT==1) then
                    call Mad_Trax (dble(send_time), dble(send_temp), send_nstep, &
                                   0, outparam%Mad_AFT_param, &
                                   Mad_AFT, dummydble17,dummydble,dummydble)
                    call correct_number(Mad_AFT)
                    allages(k) = real(Mad_AFT)
                    k = k+1
                end if
                !routines from Mad_Trax_Zircon
                if (outparam%Mad_ZFT==1) then
                    call Mad_Zirc (send_time, send_temp, send_nstep, &
                                   0, outparam%Mad_ZFT_param, &
                                   Mad_ZFT)
                    call correct_number(Mad_ZFT)
                    allages(k) = Mad_ZFT
                    k = k+1
                end if
                !routines from age_algorithms
                if (outparam%Muscovite==1) then
                    call Muscovite(dble(send_temp), dble(send_time), send_nstep, age_Muscovite)
                    call correct_number(age_Muscovite)
                    allages(k) = real(age_Muscovite)
                    k = k+1
                end if
                if (outparam%ZFT==1) then
                    call ZFT(dble(send_temp), dble(send_time), send_nstep, age_ZFT)
                    call correct_number(age_ZFT)
                    allages(k) = real(age_ZFT)
                    k = k+1
                end if
                if (outparam%Apatite_U_Th_Pb==1) then
                    call Apatite_U_Th_Pb(dble(send_temp), dble(send_time), send_nstep, age_Apatite_U_Th_Pb)
                    call correct_number(age_Apatite_U_Th_Pb)
                    allages(k) = real(age_Apatite_U_Th_Pb)
                    k = k+1
                end if
                if (outparam%Biotite==1) then
                    call Biotite(dble(send_temp), dble(send_time), send_nstep, age_Biotite)
                    call correct_number(age_Biotite)
                    allages(k) = real(age_Biotite)
                    k = k+1
                end if
                if (outparam%Rutile_U_Pb==1) then
                    call Rutile_U_Pb(dble(send_temp), dble(send_time), send_nstep, age_Rutile_U_Pb)
                    call correct_number(age_Rutile_U_Pb)
                    allages(k) = real(age_Rutile_U_Pb)
                    k = k+1
                end if
                if (outparam%Titanite_U_Pb==1) then
                    call Titanite_U_Pb(dble(send_temp), dble(send_time), send_nstep, age_Titanite_U_Pb)
                    call correct_number(age_Titanite_U_Pb)
                    allages(k) = real(age_Titanite_U_Pb)
                    k = k+1
                end if
                if (outparam%Zircon_U_Pb==1) then
                    call Zircon_U_Pb(dble(send_temp), dble(send_time), send_nstep, age_Zircon_U_Pb)
                    call correct_number(age_Zircon_U_Pb)
                    allages(k) = real(age_Zircon_U_Pb)
                    k = k+1
                end if

                !send results
                !write (*,*) iproc, ': send results of track', itrack, 'length = ',outparam%nages
                call MPI_SEND (allages,outparam%nages,MPI_REAL,0,itrack, MPI_COMM_WORLD,ierr)
                 
                deallocate (send_time)
                deallocate (send_temp)

            end do agecalc
        end if workshare
    else multiproc
        !Do calculation only on one processor
        !allocate age arrays
        if (outparam%Mad_AHe_std==1)     allocate (trcl%Mad_AHe_std(trcl%ntracks))
        if (outparam%Mad_AHe_small==1)   allocate (trcl%Mad_AHe_small(trcl%ntracks))
        if (outparam%Mad_AHe_medium==1)  allocate (trcl%Mad_AHe_medium(trcl%ntracks))
        if (outparam%Mad_AHe_large==1)   allocate (trcl%Mad_AHe_large(trcl%ntracks))
        if (outparam%Mad_AHe_lowdam==1)  allocate (trcl%Mad_AHe_lowdam(trcl%ntracks))
        if (outparam%Mad_AHe_moddam==1)  allocate (trcl%Mad_AHe_moddam(trcl%ntracks))
        if (outparam%Mad_AHe_hidam==1)   allocate (trcl%Mad_AHe_hidam(trcl%ntracks))
        if (outparam%Mad_ZHe==1)         allocate (trcl%Mad_ZHe(trcl%ntracks))
        if (outparam%Mad_ArKFs==1)       allocate (trcl%Mad_ArKFs(trcl%ntracks))
        if (outparam%Mad_ArBio==1)       allocate (trcl%Mad_ArBio(trcl%ntracks))
        if (outparam%Mad_ArMusc==1)      allocate (trcl%Mad_ArMusc(trcl%ntracks))
        if (outparam%Mad_ArHorn==1)      allocate (trcl%Mad_ArHorn(trcl%ntracks))
        if (outparam%Mad_Ti_UTh_He==1)   allocate (trcl%Mad_Ti_UTh_He(trcl%ntracks))
        if (outparam%Mad_AFT==1)         allocate (trcl%Mad_AFT(trcl%ntracks))
        if (outparam%Mad_ZFT==1)         allocate (trcl%Mad_ZFT(trcl%ntracks))
        if (outparam%Muscovite==1)       allocate (trcl%Muscovite(trcl%ntracks))
        if (outparam%ZFT==1)             allocate (trcl%ZFT(trcl%ntracks))
        if (outparam%Apatite_U_Th_Pb==1) allocate (trcl%Apatite_U_Th_Pb(trcl%ntracks))
        if (outparam%Biotite==1)         allocate (trcl%Biotite(trcl%ntracks))
        if (outparam%Rutile_U_Pb==1)     allocate (trcl%Rutile_U_Pb(trcl%ntracks))
        if (outparam%Titanite_U_Pb==1)   allocate (trcl%Titanite_U_Pb(trcl%ntracks))
        if (outparam%Zircon_U_Pb==1)     allocate (trcl%Zircon_U_Pb(trcl%ntracks))     

        do i = 1,trcl%ntracks
            !calculate temp and time array lenghts for given track
            endstep = min(1,trcl%tracks(i,6)-trcl%tracks(i,3)) 
            send_nstep = 2 + trcl%tracks(i,4)+endstep

            allocate (send_time(send_nstep))
            allocate (send_temp(send_nstep))

            !pad send-arrays at beginning, part 1
            send_time(3:send_nstep-endstep) = &
                trcl%time(trcl%tracks(i,2):trcl%tracks(i,3))
            send_time(1:2) = (/0.0,0.0/)

            send_temp(3:send_nstep-endstep) = trcl%temp(trcl%tracks(i,1), &
                trcl%tracks(i,2):trcl%tracks(i,3))
            send_temp(1) = send_temp(3)
            send_temp(2) = send_temp(3)
            
            !pad send arrays at end if necessary
            if (endstep>0) then
                send_time(send_nstep) = trcl%time(trcl%tracks(i,6))
                send_temp(send_nstep) = send_temp(send_nstep-1)
            end if

            do l=1,send_nstep
                !Calculate relative Age in Ma
                send_time(l) = send_time(send_nstep) - send_time(l)
                send_time(l) = send_time(l)*lref/veloref 
                !scale temperature
                send_temp(l) = send_temp(l)*tempscale
            end do

            !pad age array at the beginning, part 2
            send_time(1) = outparam%pad_age
            send_time(2) = 0.5*(send_time(1)+send_time(3))

            !routines from Mad_He
            if (sum(Mad_He_inputs)>0) then
                call Mad_He (dble(send_time), dble(send_temp), send_nstep, Mad_He_ages,Mad_He_inputs,21)
                if (outparam%Mad_AHe_std==1) then
                    Mad_AHe_std = Mad_He_ages(1)
                    call correct_number(Mad_AHe_std)
                    trcl%Mad_AHe_std(i) = real(Mad_AHe_std)
                end if
                if (outparam%Mad_AHe_small==1) then
                    Mad_AHe_small = Mad_He_ages(2)
                    call correct_number(Mad_AHe_small)
                    trcl%Mad_AHe_small(i) = real(Mad_AHe_small)
                end if
                if (outparam%Mad_AHe_medium==1) then
                    Mad_AHe_medium = Mad_He_ages(3)
                    call correct_number(Mad_AHe_medium)
                    trcl%Mad_AHe_medium(i) = real(Mad_AHe_medium)
                end if
                if (outparam%Mad_AHe_large==1) then
                    Mad_AHe_large = Mad_He_ages(4)
                    call correct_number(Mad_AHe_large)
                    trcl%Mad_AHe_large(i) = real(Mad_AHe_large)
                end if
                if (outparam%Mad_AHe_lowdam==1) then
                    Mad_AHe_lowdam = Mad_He_ages(5)
                    call correct_number(Mad_AHe_lowdam)
                    trcl%Mad_AHe_lowdam(i) = real(Mad_AHe_lowdam)
                end if
                if (outparam%Mad_AHe_moddam==1) then
                    Mad_AHe_moddam = Mad_He_ages(6)
                    call correct_number(Mad_AHe_moddam)
                    trcl%Mad_AHe_moddam(i) = real(Mad_AHe_moddam)
                end if
                if (outparam%Mad_AHe_hidam==1) then
                    Mad_AHe_hidam = Mad_He_ages(7)
                    call correct_number(Mad_AHe_hidam)
                    trcl%Mad_AHe_hidam(i) = real(Mad_AHe_hidam)
                end if
                if (outparam%Mad_ZHe==1) then
                    Mad_ZHe = Mad_He_ages(8)
                    call correct_number(Mad_ZHe)
                    trcl%Mad_ZHe(i) = real(Mad_ZHe)
                end if
                if (outparam%Mad_ArKFs==1) then
                    Mad_ArKFs = Mad_He_ages(12)
                    call correct_number(Mad_ArKFs)
                    trcl%Mad_ArKFs(i) = real(Mad_ArKFs)
                end if
                if (outparam%Mad_ArBio==1) then
                    Mad_ArBio = Mad_He_ages(13)
                    call correct_number(Mad_ArBio)
                    trcl%Mad_ArBio(i) = real(Mad_ArBio)
                end if
                if (outparam%Mad_ArMusc==1) then
                    Mad_ArMusc = Mad_He_ages(14)
                    call correct_number(Mad_ArMusc)
                    trcl%Mad_ArMusc(i) = real(Mad_ArMusc)
                end if
                if (outparam%Mad_ArHorn==1) then
                    Mad_ArHorn = Mad_He_ages(15)
                    call correct_number(Mad_ArHorn)
                    trcl%Mad_ArHorn(i) = real(Mad_ArHorn)
                end if
                if (outparam%Mad_Ti_UTh_He==1) then
                    Mad_Ti_UTh_He = Mad_He_ages(21)
                    call correct_number(Mad_Ti_UTh_He)
                    trcl%Mad_Ti_UTh_He(i) = real(Mad_Ti_UTh_He)
                end if
            end if
            !routine from Mad_Trax
            if (outparam%Mad_AFT==1) then
                call Mad_Trax (dble(send_time), dble(send_temp), send_nstep, &
                               0, outparam%Mad_AFT_param, &
                               Mad_AFT, dummydble17,dummydble,dummydble)
                call correct_number(Mad_AFT)
                trcl%Mad_AFT(i) = real(Mad_AFT)
            end if
            !routine from Mad_Trax_Zircon
            if (outparam%Mad_ZFT==1) then
                call Mad_Zirc (send_time, send_temp, send_nstep, &
                               0, outparam%Mad_ZFT_param, &
                               Mad_ZFT)
                call correct_number(Mad_ZFT)
                trcl%Mad_ZFT(i) = Mad_ZFT
            end if
            !routine from age_algorithms
            if (outparam%Muscovite==1) then
                call Muscovite(dble(send_temp), dble(send_time), send_nstep, age_Muscovite)
                call correct_number(age_Muscovite)
                trcl%Muscovite(i) = real(age_Muscovite)
            end if
            if (outparam%ZFT==1) then
                call ZFT(dble(send_temp), dble(send_time), send_nstep, age_ZFT)
                call correct_number(age_ZFT)
                trcl%ZFT(i) = real(age_ZFT)
            end if
            if (outparam%Apatite_U_Th_Pb==1) then
                call Apatite_U_Th_Pb(dble(send_temp), dble(send_time), send_nstep, age_Apatite_U_Th_Pb)
                call correct_number(age_Apatite_U_Th_Pb)
                trcl%Apatite_U_Th_Pb(i) = real(age_Apatite_U_Th_Pb)
            end if
            if (outparam%Biotite==1) then
                call Biotite(dble(send_temp), dble(send_time), send_nstep, age_Biotite)
                call correct_number(age_Biotite)
                trcl%Biotite(i) = real(age_Biotite)
            end if
            if (outparam%Rutile_U_Pb==1) then
                call Rutile_U_Pb(dble(send_temp), dble(send_time), send_nstep, age_Rutile_U_Pb)
                call correct_number(age_Rutile_U_Pb)
                trcl%Rutile_U_Pb(i) = real(age_Rutile_U_Pb)
            end if
            if (outparam%Titanite_U_Pb==1) then
                call Titanite_U_Pb(dble(send_temp), dble(send_time), send_nstep, age_Titanite_U_Pb)
                call correct_number(age_Titanite_U_Pb)
                trcl%Titanite_U_Pb(i) = real(age_Titanite_U_Pb)
            end if
            if (outparam%Zircon_U_Pb==1) then
                call Zircon_U_Pb(dble(send_temp), dble(send_time), send_nstep, age_Zircon_U_Pb)
                call correct_number(age_Zircon_U_Pb)
                trcl%Zircon_U_Pb(i) = real(age_Zircon_U_Pb)
            end if

            deallocate (send_time)
            deallocate (send_temp)
        end do
    end if multiproc
end if ages

!Rest done by iproc 0
iproc0: if (iproc==0) then

!------------------------------------------------------------------------------|
!    Scaling Values to natural units
!------------------------------------------------------------------------------|
!check what age prediction needs...
!units of scaling parameters:
!lref: km
!veloref: mm/yr = km/Myr
!rhoref: kg/m^3
!tempscale: °C

if (outparam%scaled==1) then
    do k=1,inparam%nstep 
        do i=1,trcl%np
            trcl%x(i,k) = trcl%x(i,k)*lref !pos in km
            trcl%y(i,k) = trcl%y(i,k)*lref
            trcl%z(i,k) = trcl%z(i,k)*lref
            trcl%temp(i,k) = trcl%temp(i,k)*tempscale !temp in C
            trcl%press(i,k) = trcl%press(i,k)*1.0e-3*9.81*rhoref*lref !pressure in MPa
        end do
        trcl%time(k) = trcl%time(k)*lref/veloref !time in Myr
    end do
end if

!------------------------------------------------------------------------------|
!    Calculating Track Output values
!------------------------------------------------------------------------------|

!track values
velo: if (outparam%velo==1) then
    allocate (trcl%velo(trcl%np,inparam%nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%velo')
    do i=1,trcl%np
            trcl%velo(i,1) = sqrt( & 
                (trcl%x(i,2)-trcl%x(i,1))**2+(trcl%y(i,2)-trcl%y(i,1))**2+(trcl%z(i,2)-trcl%z(i,1))**2) &
                /(trcl%time(2)-trcl%time(1))
    end do
    do i=1,trcl%np
        do k=2,inparam%nstep
            trcl%velo(i,k) = sqrt( & 
                (trcl%x(i,k)-trcl%x(i,k-1))**2+(trcl%y(i,k)-trcl%y(i,k-1))**2+(trcl%z(i,k)-trcl%z(i,k-1))**2) &
                /(trcl%time(k)-trcl%time(k-1))
        end do
    end do
end if velo

veloz: if (outparam%veloz==1) then
    allocate (trcl%veloz(trcl%np,inparam%nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%veloz')
    do i=1,trcl%np
        trcl%veloz(i,1) = (trcl%z(i,2)-trcl%z(i,1))/(trcl%time(2)-trcl%time(1))
    end do
    do i=1,trcl%np
        do k=2,inparam%nstep
            trcl%veloz(i,k) = (trcl%z(i,k)-trcl%z(i,k-1))/(trcl%time(k)-trcl%time(k-1))
        end do
    end do
end if veloz

!------------------------------------------------------------------------------|
!    Checking for NaNs and variable ranges
!------------------------------------------------------------------------------|

do k=1,inparam%nstep
    call correct_number(trcl%time(k))
    do i=1,trcl%np
        call correct_number(trcl%temp(i,k))
        call correct_number(trcl%press(i,k))
        if (outparam%velo==1) call correct_number(trcl%velo(i,k))
        if (outparam%veloz==1) call correct_number(trcl%veloz(i,k))
    end do
end do

!------------------------------------------------------------------------------|
!    Interpolating output onto Top Surface
!------------------------------------------------------------------------------|

allocate (surface(outparam%nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc surface')
if (btest(outparam%mode, 0) .AND. trcl%ntracks>0) then
    call interpolate_ages (trcl,surface,inparam,outparam)
end if

!Add smoothing?

!------------------------------------------------------------------------------|
!    Writing Output Files
!------------------------------------------------------------------------------|

if (outparam%outformat == 1) then
    call write_post_track (trcl, surface, inparam, outparam)
else if (outparam%outformat == 2) then
    call write_post_track_tecplot (trcl, surface, outparam)
else
    write (*,*) "ERROR: output format not allowed"
end if

!------------------------------------------------------------------------------|
!    Deallocate arrays
!------------------------------------------------------------------------------|

deallocate (trcl%x)   
deallocate (trcl%y)
deallocate (trcl%z)
deallocate (trcl%temp)
deallocate (trcl%press)
deallocate (trcl%step)
deallocate (trcl%time)
if (trcl%ntracks>0) then
    deallocate (trcl%tracks)
    if (btest(outparam%mode, 0)) then
        if (outparam%Mad_AHe_std==1)     deallocate (trcl%Mad_AHe_std)
        if (outparam%Mad_AHe_small==1)   deallocate (trcl%Mad_AHe_small)
        if (outparam%Mad_AHe_medium==1)  deallocate (trcl%Mad_AHe_medium)
        if (outparam%Mad_AHe_large==1)   deallocate (trcl%Mad_AHe_large)
        if (outparam%Mad_AHe_lowdam==1)  deallocate (trcl%Mad_AHe_lowdam)
        if (outparam%Mad_AHe_moddam==1)  deallocate (trcl%Mad_AHe_moddam)
        if (outparam%Mad_AHe_hidam==1)   deallocate (trcl%Mad_AHe_hidam)
        if (outparam%Mad_ZHe==1)         deallocate (trcl%Mad_ZHe)
        if (outparam%Mad_ArKFs==1)       deallocate (trcl%Mad_ArKFs)
        if (outparam%Mad_ArBio==1)       deallocate (trcl%Mad_ArBio)
        if (outparam%Mad_ArMusc==1)      deallocate (trcl%Mad_ArMusc)
        if (outparam%Mad_ArHorn==1)      deallocate (trcl%Mad_ArHorn)
        if (outparam%Mad_Ti_UTh_He==1)   deallocate (trcl%Mad_Ti_UTh_He)
        if (outparam%Mad_AFT==1)         deallocate (trcl%Mad_AFT)
        if (outparam%Mad_ZFT==1)         deallocate (trcl%Mad_ZFT)
        if (outparam%Muscovite==1)       deallocate (trcl%Muscovite)
        if (outparam%ZFT==1)             deallocate (trcl%ZFT)
        if (outparam%Apatite_U_Th_Pb==1) deallocate (trcl%Apatite_U_Th_Pb)
        if (outparam%Biotite==1)         deallocate (trcl%Biotite)
        if (outparam%Rutile_U_Pb==1)     deallocate (trcl%Rutile_U_Pb)
        if (outparam%Titanite_U_Pb==1)   deallocate (trcl%Titanite_U_Pb)
        if (outparam%Zircon_U_Pb==1)     deallocate (trcl%Zircon_U_Pb)     
    end if
end if
if (outparam%velo==1) deallocate (trcl%velo)
if (outparam%veloz==1) deallocate (trcl%veloz)

deallocate (surface)

end if iproc0

deallocate (allages)

call mpi_finalize (ierr)

end program post_tracking

