module m_remove_multiple_points
contains
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         |
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              REMOVE_MULTIPLE_POINT    Aug 2015                               |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

recursive subroutine remove_multiple_points (point_list,npoint,surface,flag0,surface0,is,params,nremoved)

use definitions
!use mpi
use m_toolbox
use m_remove_bad

implicit none

include 'mpif.h'

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! This routine removes the nodes given in remove_list from the surface passed
! as argument. Triangulangulation will be repaired in the vincinity of the
! removed points and special care is taken that the convex hull is kept intact.
! Bear in mind that if flag0=0, then the surface0 arrays are deallocated,
! so all operations on these arrays have to be toggled off.

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|

integer npoint,flag0,is,nremoved
integer, dimension(npoint) :: point_list
type(sheet) :: surface,surface0
type (parameters) params

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

logical :: foundinode,inode_found,jnode_found,chnode, doublenode
logical, dimension(:), allocatable :: internal,internalp

integer :: nproc,iproc,ierr
integer :: i,j,k,counter,inode,jnode,tr_counter
integer :: irem,numrem
integer :: nneighb,new_nsurface,new_nt
integer :: ninternal,nexternal,counterint,counterext
integer :: countneighb
integer, dimension(2) :: intneighb_list
integer, dimension(200) :: neighb_list
integer, dimension(:), allocatable :: remove_list,rempoints_int,rempoints_ext,new_ch
integer, dimension(:,:), allocatable :: new_icon

double precision :: eps, dist, distt
double precision :: xinode,yinode,zinode, xjnode,yjnode,zjnode

double precision, dimension(:), allocatable :: new_x,new_y,new_z
double precision, dimension(:), allocatable :: new_xn,new_yn,new_zn
double precision, dimension(:), allocatable :: new_x0,new_y0,new_z0
double precision, dimension(:), allocatable :: new_xn0,new_yn0,new_zn0

!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

eps = 1.d-8

!===============================================================================
!=====[Check for points on hull that would be removed]==========================
!===============================================================================

numrem = 0
doublenode = .false.
do i=1,npoint
    !do not remove corner points or same point twice
    inode = point_list(i)
    if ( .NOT. btest(surface%convexhull(inode), 0)) then
        if (count(point_list == inode) > 1) then
            doublenode = .true.
        else
            numrem = numrem + 1
        end if
    end if
end do



if (numrem==npoint) then
    !just copy point_list to remove_list
    allocate (remove_list(numrem))
    remove_list = point_list
else
    !Disable removal of corner nodes
    if (iproc==0) then
        write (*,'(a,i6,a,i2,a)') 'WARNING: ', npoint-numrem,' illegal nodes of surface', is, ' were marked for removal'
        write (*,*) numrem, 'points remain to be removed'
        write (8,'(a,i6,a,i2,a)') 'WARNING: ', npoint-numrem,' illegal nodes of surface', is, ' were marked for removal'
        write (8,*) numrem, 'points remain to be removed'
    end if

    if (numrem > 0) allocate (remove_list(numrem))

    irem = 0
    do i = 1,npoint
        inode = point_list(i)
        if ( (.NOT. btest(surface%convexhull(inode), 0)) .AND. &
             (count(point_list == inode) == 1)) then
            irem = irem+1
            remove_list(irem) = inode
        else
            if (iproc==0) then
                if ( btest(surface%convexhull(inode), 0) ) then
                    write(8,*) 'Not removing corner point ', inode, ' from convex hull'
                    write(8,'(a,3f10.6)') 'position: ', surface%x(inode), &
                        surface%y(inode), surface%z(inode)
                else if (count(point_list(1:i) == inode) > 1) then
                    write(8,*) 'Point ',inode,' multiple times in remove_list'
                    write(8,'(a,3f10.6)') 'position: ', surface%x(inode), &
                        surface%y(inode), surface%z(inode)
                end if
            end if
        end if
    end do

    !start anew with reduced list
    if (numrem > 0) then
        call remove_multiple_points(remove_list, irem, surface, flag0, surface0, is, params, nremoved)
        deallocate (remove_list)
    end if

    if (doublenode) then
        call remove_bad_triangles (surface, surface0, flag0, is, params, 1, nremoved)
    end if

    ! If no node has been removed until here, try to remove the duplicate nodes regularly
    if (nremoved == 0) then
        write (8,*) "Remove the following points in second round:"
        allocate (remove_list(npoint))
        irem = 0
        do i = 1,npoint
            inode = point_list(i)
            if ( (.NOT. btest(surface%convexhull(inode), 0)) .AND. &
                 (count(point_list(1:i) == inode) == 1)) then
                irem = irem+1
                remove_list(irem) = inode
                write (8,*) inode
            end if
        end do

        call remove_multiple_points(remove_list(1:irem), irem, surface, flag0, surface0, is, params, nremoved)
        deallocate (remove_list)
    end if
    return
end if

!===============================================================================
!=====[Check for internal points and restart if necessary]======================
!===============================================================================
allocate (internalp(numrem))
internalp = .true.
allocate (internal(numrem))
internal = .true.

checkinternal: do irem=1+iproc,numrem,nproc
    inode = remove_list(irem)
    countneighb = 0
    intneighb_list = 0
    checkneighb: do i=1,surface%nt
        !for each triangle, check for inode
        foundinode = .false.
        do j=1,3
            if (surface%icon(j,i)==inode) then
                foundinode = .true.
                exit
            end if
        end do
        if (foundinode) then
            !Check if point has valid neigbours, must have at least one
            !accept only convex hull nodes as valid neighbours to convex hull
            !nodes
            do j=1,3
                jnode = surface%icon(j,i)
                if (count(remove_list == jnode) == 0 .AND. &
                  .NOT. ((surface%convexhull(inode) > 0) .AND. (surface%convexhull(jnode) == 0))) then
                    internalp(irem) = .false.
                    cycle checkinternal
                end if
            end do
        end if
    end do checkneighb
end do checkinternal

call mpi_allreduce(internalp,internal,numrem,MPI_LOGICAL,MPI_LAND,MPI_COMM_WORLD,ierr)
deallocate (internalp)


!If some points have only neighbours that will be removed (internal points),
!remove these points first, then work on the rest. remove_list is split and
!adjusted for that.


ninternal = count(internal)
if (ninternal > 0) then
    if (ninternal < numrem) then
        if (iproc == 0) then
            write (8,*) '========================================================================'
            write (8,*) 'Internal nodes in remove_mutliple_points found!'
            write (8,*) 'Remove procedure will be split to remove those nodes first'
            write (8,*) '------------------------------------------------------------------------'
            write (8,*) 'Removing Internal nodes:'
        end if

        !split into internal and external points
        nexternal = numrem - ninternal
        allocate (rempoints_int(ninternal))
        allocate (rempoints_ext(nexternal))

        counterint = 0
        counterext = 0

        do irem = 1,numrem
            if (internal(irem)) then
                counterint = counterint +1
                rempoints_int(counterint) = remove_list(irem)
            else
                counterext = counterext +1
                rempoints_ext(counterext) = remove_list(irem)
            end if
        end do

        !adjust index of external points
        do irem = 1,nexternal
            inode = rempoints_ext(irem)
            rempoints_ext(irem) = inode - count(rempoints_int < inode)
        end do

        !remove independently
        call remove_multiple_points(rempoints_int, ninternal, surface, flag0, surface0, is, params, nremoved)

        if (iproc == 0) then
            write (8,*) '------------------------------------------------------------------------'
            write (8,*) 'Removing External nodes:'
        end if

        call remove_multiple_points(rempoints_ext, nexternal, surface, flag0, surface0, is, params, nremoved)

        if (iproc == 0) then
            write (8,*) '========================================================================'
        end if

        deallocate (internal)
        deallocate (rempoints_int, rempoints_ext)
        deallocate (remove_list)
        return
    else
        !ninternal == numrem -> points are bad nodes, purge from grid
        !not elegant, but hope it works
        if (iproc == 0) then
            write (8,*) '========================================================================'
            write (8,*) 'Bad nodes in remove_mutliple_points found!'
            write (8,*) '------------------------------------------------------------------------'
            write (8,*) 'Call remove_bad_triangles'
        end if
        call remove_bad_triangles (surface, surface0, flag0, is, params, 1, nremoved)
        if (iproc == 0) then
            write (8,*) '========================================================================'
        end if

        deallocate (remove_list)
        return
    end if
else
    if (iproc==0) then
        write (8,*) '------------------------------------------------------------------------'
        write (8,*) 'Removing the following points at positions:'
    end if
end if

!===============================================================================
!=====[Find neighbours and stitch icon]=========================================
!===============================================================================

remove: do irem=1,numrem
    inode = remove_list(irem) !current point to be removed
    !check for removal of convex hull nodes
    chnode = (surface%convexhull(inode) > 0)

    if (iproc==0) then
        if (chnode) then
            write (8, '(a,i6,a,3f10.6)') 'ConvHull: ', inode, ': ', surface%x(inode), &
                surface%y(inode), surface%z(inode)
        else
            write (8, '(i6,a,3f10.6)') inode, ': ', surface%x(inode), surface%y(inode), &
                surface%z(inode)
            end if
    end if

    !---------------------------------------------------------------------------
    ! Build list of neigbouring nodes
    !---------------------------------------------------------------------------
    nneighb = 0 !counting number of points for current inode
    neighb_list = 0 !list of points

    findneighb: do i=1,surface%nt
        !for each triangle, check for inode
        foundinode = .false.
        do j=1,3
            if (surface%icon(j,i)==inode) then
                foundinode = .true.
                exit
            end if
        end do
        if (foundinode) then
            !add other points to neighbour list
            !accept only convex hull as convex hull neighbours
            do j=1,3
                jnode = surface%icon(j,i)
                if (count(neighb_list == surface%icon(j,i)) == 0 .AND. &
                  count(remove_list == surface%icon(j,i)) == 0 .AND. &
                  .NOT. (chnode .AND. (surface%convexhull(jnode) == 0))) then
                    nneighb = nneighb+1
                    neighb_list(nneighb) = surface%icon(j,i)
                end if
            end do
        end if
    end do findneighb

    !if (iproc==0) then
        !write (8, *) 'neighbours and positions:'
        !do i = 1,nneighb
            !j = neighb_list(i)
            !write (8, '(i6,a,3f10.6)') j, ': ', surface%x(j), surface%y(j), &
            !    surface%z(j)
        !end do
    !end if

    !-------------------------------------------------------------------------------
    !     find vertex closest to inode
    !-------------------------------------------------------------------------------

    dist=10.d0
    do i=1,nneighb
       j=neighb_list(i)
       xinode=surface%x(inode)
       yinode=surface%y(inode)
       zinode=surface%z(inode)
       xjnode=surface%x(j)
       yjnode=surface%y(j)
       zjnode=surface%z(j)
       distt=sqrt((xinode-xjnode)**2+(yinode-yjnode)**2+(zinode-zjnode)**2)
       if (inode/=j .and. distt<dist) then
          jnode=j
          dist=distt
       end if

    end do

    !if (iproc.eq.0) then
        !write(8,*) 'Closest neighbour: ',jnode, 'at position:'
        !write(8,'(3f12.8)') surface%x(jnode),surface%y(jnode),surface%z(jnode)
        !write(8,*) ''
    !end if

    !-------------------------------------------------------------------------------
    ! carry out point removal
    !-------------------------------------------------------------------------------

    new_nsurface=surface%nsurface-1
    if (chnode) then
        new_nt=surface%nt-1
    else
        new_nt=surface%nt-2
    end if

    allocate(new_icon(3,new_nt))
    allocate(new_x(new_nsurface))
    allocate(new_y(new_nsurface))
    allocate(new_z(new_nsurface))
    allocate(new_xn(new_nsurface))
    allocate(new_yn(new_nsurface))
    allocate(new_zn(new_nsurface))
    allocate(new_ch(new_nsurface))

    if (flag0==1) then
    allocate(new_x0(new_nsurface))
    allocate(new_y0(new_nsurface))
    allocate(new_z0(new_nsurface))
    allocate(new_xn0(new_nsurface))
    allocate(new_yn0(new_nsurface))
    allocate(new_zn0(new_nsurface))
    end if

    !=====[update icon]=====

    tr_counter=0
    do i=1,surface%nt
       inode_found=.false.
       jnode_found=.false.
       do k=1,3
          if (surface%icon(k,i)==inode) inode_found=.true.
          if (surface%icon(k,i)==jnode) jnode_found=.true.
       end do
       if (.not.(inode_found .and. jnode_found)) then
          tr_counter=tr_counter+1
          new_icon(1:3,tr_counter)=surface%icon(1:3,i)
          do k=1,3
             if (new_icon(k,tr_counter)==inode) new_icon(k,tr_counter)=jnode
             if (new_icon(k,tr_counter)>inode) new_icon(k,tr_counter)=new_icon(k,tr_counter)-1
          end do
       end if
    end do

    !=====[update coordinates]=====

    counter=0
    do i=1,surface%nsurface
       if (i/=inode) then
          counter=counter+1

          new_x  (counter)=surface%x(i)
          new_y  (counter)=surface%y(i)
          new_z  (counter)=surface%z(i)
          new_xn (counter)=surface%xn(i)
          new_yn (counter)=surface%yn(i)
          new_zn (counter)=surface%zn(i)
          new_ch (counter)=surface%convexhull(i)

          if (flag0==1) then
          new_x0 (counter)=surface0%x(i)
          new_y0 (counter)=surface0%y(i)
          new_z0 (counter)=surface0%z(i)
          new_xn0(counter)=surface0%xn(i)
          new_yn0(counter)=surface0%yn(i)
          new_zn0(counter)=surface0%zn(i)
          end if

       end if
    end do

    deallocate(surface%x)
    deallocate(surface%y)
    deallocate(surface%z)
    deallocate(surface%xn)
    deallocate(surface%yn)
    deallocate(surface%zn)
    deallocate(surface%icon)
    deallocate(surface%convexhull)

    if (flag0==1) then
    deallocate(surface0%x)
    deallocate(surface0%y)
    deallocate(surface0%z)
    deallocate(surface0%xn)
    deallocate(surface0%yn)
    deallocate(surface0%zn)
    end if

    surface%nt=new_nt
    surface%nsurface=new_nsurface

    allocate(surface%x(surface%nsurface))
    allocate(surface%y(surface%nsurface))
    allocate(surface%z(surface%nsurface))
    allocate(surface%xn(surface%nsurface))
    allocate(surface%yn(surface%nsurface))
    allocate(surface%zn(surface%nsurface))
    allocate(surface%icon(3,surface%nt))
    allocate(surface%convexhull(surface%nsurface))

    if (flag0==1) then
    surface0%nsurface=new_nsurface
    allocate(surface0%x(surface0%nsurface))
    allocate(surface0%y(surface0%nsurface))
    allocate(surface0%z(surface0%nsurface))
    allocate(surface0%xn(surface0%nsurface))
    allocate(surface0%yn(surface0%nsurface))
    allocate(surface0%zn(surface0%nsurface))
    end if

    surface%x=new_x
    surface%y=new_y
    surface%z=new_z
    surface%xn=new_xn
    surface%yn=new_yn
    surface%zn=new_zn
    surface%icon=new_icon
    surface%convexhull=new_ch

    if (flag0==1) then
    surface0%x=new_x0
    surface0%y=new_y0
    surface0%z=new_z0
    surface0%xn=new_xn0
    surface0%yn=new_yn0
    surface0%zn=new_zn0
    end if

    deallocate(new_x)
    deallocate(new_y)
    deallocate(new_z)
    deallocate(new_icon)
    deallocate(new_xn)
    deallocate(new_yn)
    deallocate(new_zn)
    deallocate(new_ch)

    if (flag0==1) then
    deallocate(new_x0)
    deallocate(new_y0)
    deallocate(new_z0)
    deallocate(new_xn0)
    deallocate(new_yn0)
    deallocate(new_zn0)
    end if

    !---------------------------------------------------------------------------
    ! Adjust remove_list for next loop step
    !---------------------------------------------------------------------------
    remove_list(irem) = 0
    do i = irem+1,numrem
        if (remove_list(i) > inode) then
            remove_list(i) = remove_list(i) - 1
        end if
    end do

end do remove
nremoved = nremoved + numrem

if (iproc == 0) write (8,*) '------------------------------------------------------------------------'
deallocate (remove_list)

return

end subroutine remove_multiple_points




end module m_remove_multiple_points
