module m_update_tracking_cloud
contains
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
!                                                                              |
!              UPDATE_TRCLOUD              Jan. 2015                           |
!                                                                              |
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

subroutine update_trcloud (trcl,ov,osolve,params,istep,current_time) 

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! this file contains the operations done to 'groom' the tracking cloud - check 
! the position of particles, save surfaced ones, discard others and inject new 
! ones.
! In the first run, the cloud is created.

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
      
use definitions
!use mpi
use m_toolbox
use m_octree
use m_c_cloud

implicit none

include 'mpif.h'

type (tracking_cloud) trcl
type (octreev) ov
type (octreesolve) osolve
type (parameters) params
integer istep
double precision current_time

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

logical,dimension(:),allocatable::indomain,indomainp
integer iproc,nproc,ierr,err,iter
integer i,j,k,gridstep,tracksi
integer ntrpx,ntrpy,ntrpz
integer nsurfd,nreset
integer,dimension(:,:),allocatable::trackswap
double precision u,v,w,rat
double precision x0,y0,z0,xi,yi,zi,x,y,z
double precision,dimension(:),allocatable::xbuf,ybuf,zbuf

double precision dx,dy,dz
double precision,dimension(:),allocatable::lsf0,temp,press,lsf
double precision, dimension(5) :: sideweights
double precision, dimension(3) :: inposition

character(4) cistep
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)

!write (*,'(i2,a,i6)') iproc,': Starting update_trcloud, np = ',trcl%np

!------------------------------------------------------------------------------|
! Creating Tracking Cloud at first timestep
!------------------------------------------------------------------------------|
if (trcl%np == 0) then
!if not loading from restart file, create tracking particles according to 
!specifications in input file
    
    ntrpx = params%ntrpx
    ntrpy = params%ntrpy
    ntrpz = params%ntrpz

    injection: if (params%gridtrack) then
    !Regularly gridded tracking cloud
        !simple cubic grid
        dx = (params%trboxxmax-params%trboxxmin)/(1.d0*(ntrpx-1))
        dy = (params%trboxymax-params%trboxymin)/(1.d0*(ntrpy-1))
        dz = (params%trboxzmax-params%trboxzmin)/(1.d0*(ntrpz-1))

        !create tracking particles
        gridstep = 1
        do i = 0,ntrpx-1
            do j = 0,ntrpy-1
                do k = 0,ntrpz-1
                    trcl%x(gridstep) = params%trboxxmin+i*dx
                    trcl%y(gridstep) = params%trboxymin+j*dy
                    trcl%z(gridstep) = params%trboxzmin+k*dz
                    gridstep = gridstep+1
                end do
            end do
        end do

        if (params%trgrid_centerpt) then
            do i = 1,ntrpx-1
                do j = 1,ntrpy-1
                    do k = 1,ntrpz-1
                        trcl%x(gridstep) = params%trboxxmin+(i-0.5d0)*dx
                        trcl%y(gridstep) = params%trboxymin+(j-0.5d0)*dy
                        trcl%z(gridstep) = params%trboxzmin+(k-0.5d0)*dz
                        gridstep = gridstep+1
                    end do
                end do
            end do
        end if

        !set number of tracking particles
        trcl%np = params%ntrp

    else injection
    !Randomly distributed tracking cloud.
    !random_number is thread-safe, i.e. give same results on all processors
        call random_number (trcl%x)
        call random_number (trcl%y)
        call random_number (trcl%z)
        do i = 1,params%ntrp
            trcl%x(i) = params%trboxxmin+(params%trboxxmax-params%trboxxmin)*trcl%x(i)
            trcl%y(i) = params%trboxymin+(params%trboxymax-params%trboxymin)*trcl%y(i)
            trcl%z(i) = params%trboxzmin+(params%trboxzmax-params%trboxzmin)*trcl%z(i)
        end do
        !set number of tracking particlse
        trcl%np = params%ntrp

    end if injection
end if

!------------------------------------------------------------------------------|
! Interpolate values from osolve onto tracking particles                       |
! -- taken form update_cloud_fields.f90                                        |
!------------------------------------------------------------------------------|

!write (*,'(i2,a,i6)') iproc,': start interpolation, np = ',trcl%np

!Get upper surface level set function
allocate (lsf(osolve%nnode),stat=err) 
if (err.ne.0) call stop_run ('Error alloc lsf in update_trcloud')
if (osolve%nlsf.ne.0) then
    lsf=osolve%lsf(1:osolve%nnode,1)
else
    lsf=0.d0
end if

!allocate temporary arrays
allocate (lsf0(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('Error alloc lsf0 in update_trcloud')
allocate (temp(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('Error alloc temp in update_trcloud')
allocate (press(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('Error alloc press in update_trcloud')

!Set to 0 here because allreduce sums over all partly calcualted arrays after calculation
lsf0=0.d0
temp=0.d0
press=0.d0

!interpolate values
do k=1+iproc,trcl%np,nproc
   call octree_interpolate (osolve%octree,osolve%icon, &
                                 trcl%x(k),trcl%y(k),trcl%z(k), &
                                 lsf,lsf0(k), &
                                 ov%temp,temp(k), &
                                 ov%temporary_nodal_pressure,press(k)) 
enddo

call mpi_allreduce (lsf0, trcl%lsf0, trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
call mpi_allreduce (temp, trcl%temp, trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
call mpi_allreduce (press,trcl%press,trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)

deallocate (lsf)
deallocate (lsf0)
deallocate (temp)
deallocate (press)
!------------------------------------------------------------------------------|
! Looking for surfaced particles (lsf0>0) and saving their track extent        |
!------------------------------------------------------------------------------|

!find number of surfaced particles
nsurfd = count(trcl%lsf0 > 0.d0)

savetracks: if (nsurfd>0) then
    !increase tr acks array for newly surfaced particles
    tracksize: if (trcl%ntracks > 0) then
        !allocate swap and cop y tracks
        allocate (trackswap(trcl%ntracks,3), stat=err) 
        if (err.ne.0) call stop_run ('Error alloc trackswap in update_trcloud')
        trackswap = trcl%tracks

        !delete and reallocate tracks of increased size
        deallocate (trcl%tracks)
        allocate (trcl%tracks(trcl%ntracks+nsurfd,3), stat=err) 
        if (err.ne.0) call stop_run ('Error alloc trcl%tracks in define_trcloud')

        !copy to new tracks and deallocate swap
        trcl%tracks(1:trcl%ntracks,:) = trackswap
        deallocate (trackswap)
    else tracksize
        !if no tracks have been saved yet, trcl%ntracks should be 0, no swap is needed
        deallocate (trcl%tracks)
        allocate (trcl%tracks(nsurfd,3), stat=err) 
        if (err.ne.0) call stop_run ('Error alloc trcl%tracks in define_trcloud')
    end if tracksize

    !write new track information, reset is done in move section
    tracksi = trcl%ntracks+1
    do i = 1,trcl%np
        if (trcl%lsf0(i) > 0) then 
            trcl%tracks(tracksi,:)  = (/i,trcl%trackstart(i),istep/)
            tracksi = tracksi+1
        end if
    end do
    !update number of saved tracks
    trcl%ntracks = trcl%ntracks+nsurfd
    
end if savetracks

!------------------------------------------------------------------------------|
! Write tracking information to disk every tr_savstepth timestep after tracking|
! cloud activation.                                                            |
! All position, pressure and temperature values are saved as reals.            |
! The log of all surfaced tracks is saved.                                     |
!------------------------------------------------------------------------------|

!write (*,'(i2,a6)') iproc,': savetracks done, write to disk'

if (iproc.eq.0 .AND. mod(istep-params%tr_activation,params%tr_savstep)==0) then
    call int_to_char (cistep,4,istep)
    open (20,file='TROUT/tracking_t'//cistep//'.bin',status='replace',form='unformatted')

    !information on array size and time step 
    write (20) trcl%np,trcl%ntracks,istep,real(current_time)
 
    !scaling information
    write (20) real(params%length_scale),real(params%velocity_scale),real(params%density_scale),real(params%tempscale)
 
    !write value arrays
    write (20) real(trcl%x)
    write (20) real(trcl%y)
    write (20) real(trcl%z*params%vex)
    write (20) real(trcl%temp)
    write (20) real(trcl%press)
 
    !write tracks of surfaced particles
    write (20) trcl%tracks

    close (20)
end if

allocate (xbuf(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('Error alloc xbuf in update_trcloud')
allocate (ybuf(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('Error alloc ybuf in update_trcloud')
allocate (zbuf(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('Error alloc zbuf in update_trcloud')
     
if (params%particle_advect_method > 1) then
    xbuf = trcl%x
    ybuf = trcl%y
    zbuf = trcl%z
    call c_move_particles (osolve%octree, trcl%np, trcl%x, trcl%y, trcl%z, &
            osolve%icon, osolve%u, osolve%v, osolve%w, &
            osolve%x, osolve%y, osolve%z, &
            params%dt, params%vex)

    ! reset surfaced particles to original position
    trcl%x = merge(trcl%x, xbuf, trcl%lsf0 <= 0.d0)
    trcl%y = merge(trcl%y, ybuf, trcl%lsf0 <= 0.d0)
    trcl%z = merge(trcl%z, zbuf, trcl%lsf0 <= 0.d0)
else
    !------------------------------------------------------------------------------|
    ! Move tracking particles                                                      |
    ! taken from move_cloud.f90                                                    |
    !------------------------------------------------------------------------------|

    !write (*,'(i2,a6)') iproc,': Start moving particles'
    xbuf=0.d0
    ybuf=0.d0
    zbuf=0.d0
    rat=1.d0-1.d-10

    do i=1+iproc,trcl%np,nproc
        if (trcl%lsf0(i) <= 0.d0) then
        !For particles not surfaced, move according to interpolation of nodal velocities
            u=0.d0
            v=0.d0
            w=0.d0
            x0=trcl%x(i)*rat
            y0=trcl%y(i)*rat
            z0=trcl%z(i)*rat
            x=x0
            y=y0
            z=z0
            !midpoint algorithm 
            do iter=1,params%niter_move
               xi=(x0+x)/2.d0
               yi=(y0+y)/2.d0
               zi=(z0+z)/2.d0
               call octree_interpolate (osolve%octree,osolve%icon, &
                                            xi,yi,zi, &
                                            osolve%u,u,osolve%v,v,osolve%w/params%vex,w)
               !Move only if particle is within domain - really necessary?
               if (trcl%x(i)*(trcl%x(i)-1.d0).lt.0.d0) x=x0+params%dt*u
               if (trcl%y(i)*(trcl%y(i)-1.d0).lt.0.d0) y=y0+params%dt*v
               if (trcl%z(i)*(trcl%z(i)-1.d0).lt.0.d0) z=z0+params%dt*w
            enddo
            xbuf(i)=x
            ybuf(i)=y
            zbuf(i)=z
        else
        !if surfaced, do not move
            xbuf(i)=trcl%x(i)
            ybuf(i)=trcl%y(i)
            zbuf(i)=trcl%z(i)
        end if
    enddo

    trcl%x=0.d0
    trcl%y=0.d0
    trcl%z=0.d0
    call mpi_allreduce (xbuf,trcl%x,trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce (ybuf,trcl%y,trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce (zbuf,trcl%z,trcl%np,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
end if

deallocate (xbuf)
deallocate (ybuf)
deallocate (zbuf)

!------------------------------------------------------------------------------|
! Reset surfaced and exited particles                                          |
!------------------------------------------------------------------------------|

!write (*,'(i2,a)') iproc,': start reset particles'

allocate (indomain(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('error in alloc indomain in update_trcloud')
allocate (indomainp(trcl%np),stat=err) 
if (err.ne.0) call stop_run ('error in alloc indomainp in update_trcloud')

indomain=.false.
indomainp=.false.

do i=1+iproc,trcl%np,nproc
   indomainp(i) = 0.0d0<trcl%x(i) .AND. trcl%x(i)<1.d0 .AND. &
       0.0d0<trcl%y(i) .AND. trcl%y(i)<1.d0 .AND. &
       0.0d0<trcl%z(i) .AND. trcl%lsf0(i)<0.d0
   !write (*,'(a,4F10.6,l2)') 'position', trcl%x(i), trcl%y(i), trcl%z(i), trcl%lsf0(i), indomainp(i)
enddo

call mpi_allreduce (indomainp,indomain,trcl%np,mpi_logical,mpi_lor,mpi_comm_world,ierr)
deallocate (indomainp)

call calculate_sideweights (sideweights, params, osolve)
!if (iproc==0) write (*, '(5f8.5)') sideweights

nreset=0
do i=1,trcl%np
    if (.NOT. indomain(i)) then
        call injectposition (inposition, sideweights, params)
        !if (iproc==0) write (*,'(a,3f8.5)') 'new particle injected at ', inposition

        trcl%x(i) = inposition(1)
        trcl%y(i) = inposition(2)
        trcl%z(i) = inposition(3)

        !reset trackstart
        trcl%trackstart(i)= istep+1
        nreset = nreset+1
    end if
end do 

if (iproc==0) write (*,'(a,i8,a,i8,a)') 'update_trcloud:',nsurfd, & 
    ' particles surfaced and in total ',nreset,' particles reset'

deallocate (indomain)

return
end subroutine update_trcloud 

subroutine calculate_sideweights (sideweights, params, osolve)

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! For tracking reinjection, particles are weighed by the total inflow across the 
! sides. This routine calculates the weights.

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
      
use definitions
use m_octree

implicit none

type (octreesolve), intent(in) :: osolve
type (parameters), intent(in) :: params
double precision, dimension(5), intent(out) :: sideweights

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

double precision, dimension(5) :: area, influx, npoints
double precision, dimension(3) :: norm, velocity
double precision :: x, y, z, dx, dy, dz
double precision :: boxxmin, boxxmax, boxymin, boxymax, boxzmin, boxzmax
double precision :: totinflux, eps

integer :: i, k, nx, ny, nz, ix, iy, iz, ystep

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

eps = 1.0d-6

boxxmin = params%trboxxmin
boxxmax = params%trboxxmax
boxymin = params%trboxymin
boxymax = params%trboxymax
boxzmin = params%trboxzmin
boxzmax = params%trboxzmax

area(1) = (boxymax-boxymin) * (boxzmax-boxzmin) * params%vex
area(2) = (boxxmax-boxxmin) * (boxzmax-boxzmin) * params%vex
area(3) = area(1)
area(4) = area(2)
area(5) = (boxxmax-boxxmin) * (boxymax-boxymin)

!Estimate influx by averaging sample points
nx = 20
ny = 20
nz = 10
dx = (boxxmax - boxxmin)/(nx*1.d0)
dy = (boxymax - boxymin)/(ny*1.d0)
dz = (boxzmax - boxzmin)/(nz*1.d0)
npoints = 0.d0
do ix = 0,nx
    do iz = 0,nz
        if (iz == 0 .OR. ix == 0 .OR. ix == nx) then
            ystep = 1
        else
            ystep = ny + 1
        end if

        do iy = 0, ny, ystep
            x = boxxmin + ix * dx
            y = boxymin + iy * dy
            z = boxzmin + iz * dz
   
            call octree_interpolate (osolve%octree,osolve%icon, &
                     x, y, z, &
                     osolve%u, velocity(1), &
                     osolve%v, velocity(2), &
                     osolve%w/params%vex, velocity(3))

            if (ix == 0) then
                norm =  (/1.d0, 0.d0, 0.d0/)
                influx(1) = influx(1) + dot_product( norm, velocity )
                npoints(1) = npoints(1) + 1.d0
            else if (ix == nx) then
                norm =  (/-1.d0, 0.d0, 0.d0/)
                influx(3) = influx(3) + dot_product( norm, velocity )
                npoints(3) = npoints(3) + 1.d0
            end if
            if (iy == 0) then
                norm =  (/0.d0, 1.d0, 0.d0/)
                influx(2) = influx(2) + dot_product( norm, velocity )
                npoints(2) = npoints(2) + 1.d0
            else if (iy == ny) then
                norm =  (/0.d0, -1.d0, 0.d0/)
                influx(4) = influx(4) + dot_product( norm, velocity )
                npoints(4) = npoints(4) + 1.d0
            end if
            if (iz == 0) then
                norm =  (/0.d0, 0.d0, 1.d0/)
                influx(5) = influx(5) + dot_product( norm, velocity )
                npoints(5) = npoints(5) + 1.d0
            end if

        end do
    end do
end do

! calculate influx as velocity*area and normalize
do k = 1,5
    influx(k) = max(0.d0, influx(k) * area(k) / npoints(k))
end do
totinflux = sum(influx)
! if only outflux at all sides
if (totinflux < eps) then
    influx(:) = area (:)
    totinflux = sum(influx)
end if
sideweights = influx / totinflux

end subroutine calculate_sideweights

subroutine injectposition ( iposition, sideweights, params)

!------------------------------------------------------------------------------|
!(((((((((((((((( Purpose of the routine  ))))))))))))))))))))))))))))))))))))))
!------------------------------------------------------------------------------|
! For tracking reinjection, particles are weighed by the total inflow across the 
! sides. This routine calculates a random inject position.

!------------------------------------------------------------------------------|
!((((((((((((((((  declaration of the subroutine arguments  ))))))))))))))))))))
!------------------------------------------------------------------------------|
      
use definitions

implicit none

type (parameters), intent(in) ::  params
double precision, dimension(5), intent(in) :: sideweights
double precision, dimension(3), intent(out) :: iposition

!------------------------------------------------------------------------------|
!(((((((((((((((( declaration of the subroutine internal variables )))))))))))))
!------------------------------------------------------------------------------|

double precision :: boxxmin, boxxmax, boxymin, boxymax, boxzmin, boxzmax
double precision :: a, csum0, csum1
integer :: i, iside
!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

boxxmin = params%trboxxmin
boxxmax = params%trboxxmax
boxymin = params%trboxymin
boxymax = params%trboxymax
boxzmin = params%trboxzmin
boxzmax = params%trboxzmax

call random_number(a)

iside = 0
csum0 = 0.d0
csum1 = 0.d0
do i = 1,5
    csum0 = csum1
    csum1 = csum1 + sideweights(i)
    if (csum0 < a .AND. a < csum1) then
        iside = i
        exit
    end if
end do
if (iside == 0) iside = 1


call random_number (iposition)
iposition = iposition * (/boxxmax-boxxmin, boxymax-boxymin, boxzmax-boxzmin/) + &
                (/boxxmin, boxymin, boxzmin/) 

select case( iside )
    case (1)
        iposition(1) = boxxmin
    case (2)
        iposition(2) = boxymin
    case (3)
        iposition(1) = boxxmax
    case (4)
        iposition(2) = boxymax
    case (5)
        iposition(3) = boxzmin 
    case default
        iposition = (/ 0.5d0, 0.5d0, 0.0d0 /)
        write (*,*) 'Error: cannot find correct box side for injection'
end select
end subroutine injectposition


!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|
end module m_update_tracking_cloud

